#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include <stdbool.h>
#include <ncurses.h>
#include <unistd.h>
#include <termios.h>
#include <limits.h>
#include <time.h>
#include <SDL2/SDL.h>
#include <SDL2/SDL_image.h>

#include "hash.h"
#include "structure.h"
#include "base.h"
#include "generation.h"

// ------------------------------------------------------------------------ //

Grid map ;

template* configs ;
array** matches ;
int n_configs = 18;

int render_distance = 4 ;

int player_x = 4;
int player_y = 4;

int player_cx = 0;
int player_cy = 0;

int zoom = 75 ;
int draw_par = 1;

int __width__ = 1200 ;
int __height__ = 800 ;

imgs digits ;
imgs letters ;

template full ;

// ------------------------------------------------------------------------ //

void one_debug() {
    matches = malloc(sizeof(array*)*4);

    matches[0] = malloc(sizeof(array));
    matches[1] = malloc(sizeof(array));
    matches[2] = malloc(sizeof(array));
    matches[3] = malloc(sizeof(array));

    matches[0][0].arr = malloc(sizeof(int));
    matches[0][0].arr[0] = 0;
    matches[0][0].len = 1;

    matches[1][0].arr = malloc(sizeof(int));
    matches[1][0].arr[0] = 0;
    matches[1][0].len = 1;

    matches[2][0].arr = malloc(sizeof(int));
    matches[2][0].arr[0] = 0;
    matches[2][0].len = 1;

    matches[3][0].arr = malloc(sizeof(int));
    matches[3][0].arr[0] = 0;
    matches[3][0].len = 1;

    configs = malloc(sizeof(template)); 

    configs[0].lines = malloc(sizeof(uint8_t)*8);
    configs[0].lines[0] = 255;
    configs[0].lines[1] = 129;
    configs[0].lines[2] = 129;
    configs[0].lines[3] = 129;
    configs[0].lines[4] = 129;
    configs[0].lines[5] = 129;
    configs[0].lines[6] = 129;
    configs[0].lines[7] = 255;

    configs[0].id = 0;
}

// ------------------------------------------------------------------------ //

bool unpack_coord(uint8_t* lines, int cx, int cy) {
    return (bool)((lines[cy]/pw(2, cx))%2);
}

bool is_compatible(chunk ch, cardinal dir, int config_id) {
    return arr_mem(matches[(int)dir][ch.chdata.id], config_id);
}

template copy(template src) {
    template new ;
    new.id = src.id ;
    new.eastsig = src.eastsig;
    new.westsig = src.westsig;
    new.lines = malloc(sizeof(uint8_t)*8);
    for(int i = 0; i < 8; i++) {
        new.lines[i] = src.lines[i];
    };
    return new;
}

void generate_chunk(int x, int y, int config_id) { 
    chunk* new = malloc(sizeof(chunk)) ;
    new->chx = x ;
    new->chy = y ;
    new->draw_id = draw_par - 2;
    if(config_id >= 0) {
        new->chdata = copy(configs[config_id]);
    } else {
        new->chdata = copy(full) ;
    };
    //printf("inserting chunk (%d, %d)[%d]\n", new->chx, new->chy, new->draw_id);
    gridInsert(map, x, y, new);
}

void print_template(int config_id) {
    for(int i = 0; i < 8; i++) {
        uint8_t rem = configs[config_id].lines[i];
        for(int j = 0; j < 8; j++) {
            if(rem%2 == 1) {
                printf("X ");
            } else {
                printf("  ");
            };
            rem = rem/2 ;
        };
        printf("(%d)\n", configs[config_id].lines[i]);
    };/*
    for(int i = 0; i < 4; i++) {
        printf("[");
        for(int j = 0; j < matches[i][config_id].len; j++) {
            printf("%d ", matches[i][config_id].arr[j]);
        }
        printf("]\n");
    };*/
    printf("\n");
}

void rotateTemplateClockWise(int config_id) {
    // 1st line become last column ... last line become 1st column
    uint8_t* new = malloc(sizeof(uint8_t)*8);

    configs[config_id].westsig = configs[config_id].lines[7];
    configs[config_id].eastsig = configs[config_id].lines[0];

    for(int i = 0; i < 8; i++) {
        new[i] = configs[config_id].lines[i];
        configs[config_id].lines[i] = 0;
    };    
    for(int i = 0; i < 8; i++) {
        uint8_t rem = new[i];
        for(int j = 0; j < 8; j++) {
            configs[config_id].lines[j] *= 2 ;
            configs[config_id].lines[j] += rem%2 ;
            rem = rem / 2 ;
        }
    };

    if(config_id >= 9 ) {
        //print_template(config_id);
    }

    if(false) {
        int* temp = matches[0][config_id].arr;
        int len0 = matches[0][config_id].len;

        matches[0][config_id].arr = matches[1][config_id].arr;
        matches[0][config_id].len = matches[1][config_id].len;

        matches[1][config_id].arr = matches[2][config_id].arr;
        matches[1][config_id].len = matches[2][config_id].len;

        matches[2][config_id].arr = matches[3][config_id].arr;
        matches[2][config_id].len = matches[3][config_id].len;

        matches[3][config_id].arr = temp;
        matches[3][config_id].len = len0;

        for(int tp = 0; tp < n_configs; tp++) {
            pop(&matches[0][tp], config_id);
            append(&matches[1][tp], config_id);

            pop(&matches[1][tp], config_id);
            append(&matches[2][tp], config_id);

            pop(&matches[2][tp], config_id);
            append(&matches[3][tp], config_id);

            pop(&matches[3][tp], config_id);
            append(&matches[0][tp], config_id);
        }
    }

    free(new);
}

bool is_compat_one(int cx, int cy, cardinal dir, int idx) {
    if(idx == -1 || !gridMem(map, cx, cy)) {
        return true;
    };
    return is_compatible(*gridGet(map, cx, cy), dir, idx);
}

bool is_compat_sig_north(int cx, int cy, int idx) {
    chunk* cp = gridGet(map, cx, cy-1);
    if(cp == NULL) {
        return true;
    } else {
        return cp->chdata.lines[7] == configs[idx].lines[0];
    }
}
bool is_compat_sig_south(int cx, int cy, int idx) {
    chunk* cp = gridGet(map, cx, cy+1);
    if(cp == NULL) {
        return true;
    } else {
        return cp->chdata.lines[0] == configs[idx].lines[7];
    }
}
bool is_compat_sig_east(int cx, int cy, int idx) {
    chunk* cp = gridGet(map, cx+1, cy);
    if(cp == NULL) {
        return true;
    } else {
        return cp->chdata.westsig == configs[idx].eastsig;
    }
}
bool is_compat_sig_west(int cx, int cy, int idx) {
    chunk* cp = gridGet(map, cx-1, cy);
    if(cp == NULL) {
        return true;
    } else {
        return cp->chdata.eastsig == configs[idx].westsig;
    }
}

void signature(template t) {
    printf("[NORTH] %d\n", t.lines[0]);
    printf("[EAST]  %d\n", t.eastsig);
    printf("[SOUTH] %d\n", t.lines[7]);
    printf("[WEST]  %d\n\n", t.westsig);
}

bool is_compat_with_spins(int cx, int cy, int idx) {
    for(int k = 0; k < 4; k++) {
        //printf("    (%d %d %d %d)\n", is_compat_one(cx-1, cy, EAST, idx), is_compat_one(cx+1, cy, WEST, idx), is_compat_one(cx, cy-1, SOUTH, idx), is_compat_one(cx, cy+1, NORTH, idx));
        //if(is_compat_one(cx-1, cy, EAST, idx) && is_compat_one(cx+1, cy, WEST, idx) && is_compat_one(cx, cy-1, SOUTH, idx) && is_compat_one(cx, cy+1, NORTH, idx)) {
        if(is_compat_sig_north(cx, cy, idx) && is_compat_sig_east(cx, cy, idx) && is_compat_sig_south(cx, cy, idx) && is_compat_sig_west(cx, cy, idx)) {
            
            /*printf("---------------------------------\n");
            signature(configs[idx]);
            if(gridGet(map, cx, cy+1) != NULL) {
                printf("[SOUTH] - %d\n", (gridGet(map, cx, cy+1)->chdata).lines[7]);
            } else {
                printf("[SOUTH] - \n");
            }
            if(gridGet(map, cx+1, cy) != NULL) {
                printf("[EAST]  - %d\n", (gridGet(map, cx+1, cy)->chdata).westsig);
            } else {
                printf("[EAST]  - \n");
            }
            if(gridGet(map, cx, cy-1) != NULL) {
                printf("[NORTH] - %d\n", (gridGet(map, cx, cy-1)->chdata).lines[0]);
            } else {
                printf("[NORTH] - \n");
            }
            if(gridGet(map, cx-1, cy) != NULL) {
                printf("[WEST]  - %d\n\n", (gridGet(map, cx-1, cy)->chdata).eastsig);
            } else {
                printf("[WEST]  - \n");
            }
            printf("---------------------------------\n");*/

            return true ;
        } else {
            rotateTemplateClockWise(idx);
        }
    };
    return false;
}

bool is_compat_with_spins_uno(int cx, int cy, int idx) {
    if(is_compat_sig_north(cx, cy, idx) && is_compat_sig_east(cx, cy, idx) && is_compat_sig_south(cx, cy, idx) && is_compat_sig_west(cx, cy, idx)) {
        return true ;
    } else {
        rotateTemplateClockWise(idx);
        return false;
    }
}

void generate_chunk_all(int cx, int cy) {
    int i = 1 + (rand()%(4*n_configs));
    int idx = 0 ;
    int turn = 0 ;
    int k = 0 ;
    while(i > 0) {
        if(turn > 8*n_configs) {
            i = 0 ;
            idx = 0 ;
            printf("failed\n");
        } else {
            if(is_compat_with_spins_uno(cx, cy, idx)) {
                i -= 1;
                if(i != 0) {
                    if(k == 3) {
                        idx = (idx+1)%n_configs;
                        k = 0 ;
                    } else {
                        k++ ;
                    };
                }
            } else {
                if(k == 3) {
                    idx = (idx+1)%n_configs;
                    turn += 1;
                    k = 0 ;
                } else {
                    k++ ;
                }
            }
        }
    };
    //printf("\ngenerating %d\n", idx);
    generate_chunk(cx, cy, idx);
}

void initialize(SDL_Renderer* renderer) {
    map = newGrid();

    import_letters(renderer);
    import_digits(renderer);

    full.lines = malloc(sizeof(uint8_t)*8);
    for(int i = 0; i < 8; i++) {
        full.lines[i] = 255 ;
    };
    full.id = -1 ;

    generate_chunk(0, 0, 1);
}

void destroy() {
    freeGrid(map);

    free_digits(digits);
    free_digits(letters);

    free(full.lines);
}

void parse_one(FILE* ptr, FILE* ptr2, int index) {
    for(int i = 0; i < 8; i++) {
        for(int j = 0; j < 8; j++) {
            int xres = get_integer_plus_align(ptr, ptr2);
            configs[index].lines[i] *= 2 ;
            configs[index].lines[i] += xres ;

            if(j == 0) {
                configs[index].westsig *= 2;
                configs[index].westsig += xres;
            } else if(j == 7) {
                configs[index].eastsig *= 2;
                configs[index].eastsig += xres;
            }
        }
    };/*
    for(int i = 0; i < 4; i++) {
        int len = count_char_in_line(ptr2, ',');
        matches[i][index].len = len ;
        matches[i][index].arr = malloc(sizeof(int)*n_configs); // just in case
        int wasted = fgetc(ptr); // always a '['
        for(int j = 0; j < len; j++) {
            int read = get_integer(ptr);
            matches[i][index].arr[j] = read ;
        };
        terminate_line(ptr);
    };*/
    
    terminate_line(ptr);
    terminate_line(ptr2);
}

void parse_configs(char* filename, int n_conf) {
    // Initializing everything //
    n_configs = n_conf ;

    matches = malloc(sizeof(array*)*4);
    // 4 cardinal directions

    matches[0] = malloc(sizeof(array)*n_conf);
    matches[1] = malloc(sizeof(array)*n_conf);
    matches[2] = malloc(sizeof(array)*n_conf);
    matches[3] = malloc(sizeof(array)*n_conf);
    // compatibilities, 1 for each template

    configs = malloc(sizeof(template)*n_conf);
    // data for each template

    for(int i = 0; i < n_conf; i++) {
        configs[i].id = i;
        configs[i].eastsig = 0 ;
        configs[i].westsig = 0 ;
        configs[i].lines = malloc(sizeof(uint8_t)*8);
        for(int j = 0; j < 8; j++) {
            configs[i].lines[j] = 0;
        }
    }

    // Now for the fun part //
    FILE* ptr = fopen(filename, "r");
    FILE* ptr2 = fopen(filename, "r");

    for(int k = 0; k < n_conf; k++) {
        parse_one(ptr, ptr2, k);
    }
}